@@ -20,6 +20,7 @@ from octue.mixins import CoolNameable
 from octue.utils.encoders import OctueJSONEncoder
 from octue.utils.exceptions import convert_exception_to_primitives
 from octue.utils.objects import get_nested_attribute
+from octue.utils.threads import RepeatingTimer
 
 
 logger = logging.getLogger(__name__)
@@ -64,6 +65,10 @@ class Service(CoolNameable):
 
         self.backend = backend
         self.run_function = run_function
+        self._raw_local_sdk_version = pkg_resources.get_distribution("octue").version
+        self._parsed_local_sdk_version = packaging.version.parse(self._raw_local_sdk_version)
+        self._record_sent_messages = False
+        self._sent_messages = []
         self._publisher = None
         self._credentials = None
         super().__init__(*args, **kwargs)
@@ -147,12 +152,13 @@ class Service(CoolNameable):
 
             subscriber.close()
 
-    def answer(self, question, answer_topic=None, timeout=30):
+    def answer(self, question, answer_topic=None, heartbeat_interval=120, timeout=30):
         """Answer a question from a parent - i.e. run the child's app on the given data and return the output values.
         Answers conform to the output values and output manifest schemas specified in the child's Twine file.
 
         :param dict|Message question:
         :param octue.cloud.pub_sub.topic.Topic|None answer_topic: provide if messages need to be sent to the parent from outside the `Service` instance (e.g. in octue.cloud.deployment.google.cloud_run.flask_app)
+        :param int|float heartbeat_interval: the time interval, in seconds, at which to send heartbeats
         :param float|None timeout: time in seconds to keep retrying sending of the answer once it has been calculated
         :raise Exception: if any exception arises during running analysis and sending its results
         :return None:
@@ -165,13 +171,28 @@ class Service(CoolNameable):
             allow_save_diagnostics_data_on_crash,
         ) = self._parse_question(question)
 
+        # Record messages sent to child for potential diagnostics.
+        if allow_save_diagnostics_data_on_crash:
+            self._record_sent_messages = True
+        else:
+            self._record_sent_messages = False
+
         topic = answer_topic or self.instantiate_answer_topic(question_uuid)
         self._send_delivery_acknowledgment(topic)
 
+        heartbeater = RepeatingTimer(
+            interval=heartbeat_interval,
+            function=self._send_heartbeat,
+            kwargs={"topic": topic},
+        )
+
+        heartbeater.daemon = True
+        heartbeater.start()
+
         try:
             if forward_logs:
                 analysis_log_handler = GooglePubSubHandler(
-                    publisher=self.publisher,
+                    message_sender=self._send_message,
                     topic=topic,
                     analysis_id=question_uuid,
                 )
@@ -179,18 +200,16 @@ class Service(CoolNameable):
                 analysis_log_handler = None
 
             if parent_sdk_version:
-                local_sdk_version = packaging.version.parse(pkg_resources.get_distribution("octue").version)
-
                 if (
-                    local_sdk_version.major != parent_sdk_version.major
-                    or local_sdk_version.minor != parent_sdk_version.minor
+                    self._parsed_local_sdk_version.major != parent_sdk_version.major
+                    or self._parsed_local_sdk_version.minor != parent_sdk_version.minor
                 ):
                     logger.warning(
                         "The parent's Octue SDK version %s may not be compatible with the local Octue SDK version %s. "
                         "Update them both to the latest version (or at least a version with the same major and minor "
                         "version numbers) if possible.",
                         parent_sdk_version,
-                        local_sdk_version,
+                        self._parsed_local_sdk_version,
                     )
 
             else:
@@ -206,6 +225,7 @@ class Service(CoolNameable):
                 analysis_log_handler=analysis_log_handler,
                 handle_monitor_message=functools.partial(self._send_monitor_message, topic=topic),
                 allow_save_diagnostics_data_on_crash=allow_save_diagnostics_data_on_crash,
+                sent_messages=self._sent_messages,
             )
 
             if analysis.output_manifest is None:
@@ -213,23 +233,22 @@ class Service(CoolNameable):
             else:
                 serialised_output_manifest = analysis.output_manifest.serialise()
 
-            self.publisher.publish(
-                topic=topic.path,
-                data=json.dumps(
-                    {
-                        "type": "result",
-                        "output_values": analysis.output_values,
-                        "output_manifest": serialised_output_manifest,
-                        "message_number": topic.messages_published,
-                    },
-                    cls=OctueJSONEncoder,
-                ).encode(),
-                retry=retry.Retry(deadline=timeout),
+            self._send_message(
+                {
+                    "type": "result",
+                    "output_values": analysis.output_values,
+                    "output_manifest": serialised_output_manifest,
+                    "message_number": topic.messages_published,
+                },
+                topic=topic,
+                timeout=timeout,
             )
-            topic.messages_published += 1
+
+            heartbeater.cancel()
             logger.info("%r answered question %r.", self, question_uuid)
 
         except BaseException as error:  # noqa
+            heartbeater.cancel()
             self.send_exception(topic, timeout)
             raise error
 
@@ -296,14 +315,12 @@ class Service(CoolNameable):
             input_manifest.use_signed_urls_for_datasets()
             serialised_input_manifest = input_manifest.serialise()
 
-        self.publisher.publish(
-            topic=question_topic.path,
-            data=json.dumps({"input_values": input_values, "input_manifest": serialised_input_manifest}).encode(),
+        self._send_message(
+            {"input_values": input_values, "input_manifest": serialised_input_manifest},
+            topic=question_topic,
             question_uuid=question_uuid,
-            forward_logs=str(int(subscribe_to_logs)),
-            octue_sdk_version=pkg_resources.get_distribution("octue").version,
-            allow_save_diagnostics_data_on_crash=str(int(allow_save_diagnostics_data_on_crash)),
-            retry=retry.Retry(deadline=timeout),
+            forward_logs=subscribe_to_logs,
+            allow_save_diagnostics_data_on_crash=allow_save_diagnostics_data_on_crash,
         )
 
         logger.info("%r asked a question %r to service %r.", self, question_uuid, unlinted_service_id)
@@ -313,18 +330,22 @@ class Service(CoolNameable):
         self,
         subscription,
         handle_monitor_message=None,
+        record_messages_to=None,
         service_name="REMOTE",
         timeout=60,
         delivery_acknowledgement_timeout=120,
+        maximum_heartbeat_interval=300,
     ):
         """Wait for an answer to a question on the given subscription, deleting the subscription and its topic once
         the answer is received.
 
         :param octue.cloud.pub_sub.subscription.Subscription subscription: the subscription for the question's answer
         :param callable|None handle_monitor_message: a function to handle monitor messages (e.g. send them to an endpoint for plotting or displaying) - this function should take a single JSON-compatible python primitive as an argument (note that this could be an array or object)
+        :param str|None record_messages_to: if given a path to a JSON file, messages received in response to the question are saved to it
         :param str service_name: a name by which to refer to the child subscribed to (used for labelling its log messages if subscribed to)
         :param float|None timeout: how long in seconds to wait for an answer before raising a `TimeoutError`
         :param float delivery_acknowledgement_timeout: how long in seconds to wait for a delivery acknowledgement before aborting
+        :param float|int maximum_heartbeat_interval: the maximum amount of time (in seconds) allowed between child heartbeats before an error is raised
         :raise TimeoutError: if the timeout is exceeded
         :raise octue.exceptions.QuestionNotDelivered: if a delivery acknowledgement is not received in time
         :return dict: dictionary containing the keys "output_values" and "output_manifest"
@@ -341,12 +362,14 @@ class Service(CoolNameable):
             subscription=subscription,
             handle_monitor_message=handle_monitor_message,
             service_name=service_name,
+            record_messages_to=record_messages_to,
         )
 
         try:
             return message_handler.handle_messages(
                 timeout=timeout,
                 delivery_acknowledgement_timeout=delivery_acknowledgement_timeout,
+                maximum_heartbeat_interval=maximum_heartbeat_interval,
             )
         finally:
             subscription.delete()
@@ -375,22 +398,18 @@ class Service(CoolNameable):
         exception = convert_exception_to_primitives()
         exception_message = f"Error in {self!r}: {exception['message']}"
 
-        self.publisher.publish(
-            topic=topic.path,
-            data=json.dumps(
-                {
-                    "type": "exception",
-                    "exception_type": exception["type"],
-                    "exception_message": exception_message,
-                    "traceback": exception["traceback"],
-                    "message_number": topic.messages_published,
-                }
-            ).encode(),
-            retry=retry.Retry(deadline=timeout),
+        self._send_message(
+            {
+                "type": "exception",
+                "exception_type": exception["type"],
+                "exception_message": exception_message,
+                "traceback": exception["traceback"],
+                "message_number": topic.messages_published,
+            },
+            topic=topic,
+            timeout=timeout,
         )
 
-        topic.messages_published += 1
-
     def _clean_service_id(self, service_id):
         """Replace forward slashes in the given service ID with dots.
 
@@ -399,28 +418,72 @@ class Service(CoolNameable):
         """
         return service_id.replace("/", ".")
 
+    def _send_message(self, message, topic, timeout=30, **attributes):
+        """Send a JSON-serialised message to the given topic with optional message attributes.
+
+        :param any message: any JSON-serialisable python primitive to send as a message
+        :param octue.cloud.pub_sub.topic.Topic topic: the Pub/Sub topic to send the message to
+        :param int|float timeout: the timeout for sending the message in seconds
+        :param attributes: key-value pairs to attach to the message - the values must be strings or bytes
+        :return None:
+        """
+        attributes["octue_sdk_version"] = self._raw_local_sdk_version
+        converted_attributes = {}
+
+        for key, value in attributes.items():
+            if isinstance(value, bool):
+                value = str(int(value))
+            converted_attributes[key] = value
+
+        self.publisher.publish(
+            topic=topic.path,
+            data=json.dumps(message, cls=OctueJSONEncoder).encode(),
+            retry=retry.Retry(deadline=timeout),
+            **converted_attributes,
+        )
+
+        topic.messages_published += 1
+
+        if self._record_sent_messages:
+            self._sent_messages.append(message)
+
     def _send_delivery_acknowledgment(self, topic, timeout=30):
         """Send an acknowledgement of question receipt to the parent.
 
-        :param octue.cloud.pub_sub.topic.Topic topic: topic to send acknowledgement to
+        :param octue.cloud.pub_sub.topic.Topic topic: topic to send the acknowledgement to
         :param float timeout: time in seconds after which to give up sending
         :return None:
         """
         logger.info("%r acknowledged receipt of question.", self)
 
-        self.publisher.publish(
-            topic=topic.path,
-            data=json.dumps(
-                {
-                    "type": "delivery_acknowledgement",
-                    "delivery_time": str(datetime.datetime.now()),
-                    "message_number": topic.messages_published,
-                }
-            ).encode(),
-            retry=retry.Retry(deadline=timeout),
+        self._send_message(
+            {
+                "type": "delivery_acknowledgement",
+                "delivery_time": str(datetime.datetime.now()),
+                "message_number": topic.messages_published,
+            },
+            topic=topic,
+            timeout=timeout,
         )
 
-        topic.messages_published += 1
+    def _send_heartbeat(self, topic, timeout=30):
+        """Send a heartbeat to the parent, indicating that the service is alive.
+
+        :param octue.cloud.pub_sub.topic.Topic topic: topic to send the heartbeat to
+        :param float timeout: time in seconds after which to give up sending
+        :return None:
+        """
+        self._send_message(
+            {
+                "type": "heartbeat",
+                "time": str(datetime.datetime.now()),
+                "message_number": topic.messages_published,
+            },
+            topic=topic,
+            timeout=timeout,
+        )
+
+        logger.debug("Heartbeat sent.")
 
     def _send_monitor_message(self, data, topic, timeout=30):
         """Send a monitor message to the parent.
@@ -432,20 +495,16 @@ class Service(CoolNameable):
         """
         logger.debug("%r sending monitor message.", self)
 
-        self.publisher.publish(
-            topic=topic.path,
-            data=json.dumps(
-                {
-                    "type": "monitor_message",
-                    "data": json.dumps(data),
-                    "message_number": topic.messages_published,
-                }
-            ).encode(),
-            retry=retry.Retry(deadline=timeout),
+        self._send_message(
+            {
+                "type": "monitor_message",
+                "data": json.dumps(data),
+                "message_number": topic.messages_published,
+            },
+            topic=topic,
+            timeout=timeout,
         )
 
-        topic.messages_published += 1
-
     def _parse_question(self, question):
         """Parse a question in the Google Cloud Pub/Sub or Google Cloud Run format.
 
